'''SbSessionEx provides a session object for working with ScienceBase GraphQL
'''
import json
import logging
from datetime import datetime
import requests
from sb3 import auth, client

class SbSessionEx:
    """SbSessionEx provides extra methods for working with ScienceBase GraphQL
    as well as authentication using Keycloak.
    """
    _env = None
    _graphql_url = None
    _realm = None
    _auth_server_url = None
    _username = None
    _token = None
    _refresh_token = None
    _client_id = None
    _token_server_uri = None
    _token_expire = None
    _token_expire_refresh = None
    _token_expire_time = None
    _is_logged_in = False

    def __init__(self, env=None):
        self._env = env
        self._logging = logging
        self._auth_server_url = "https://www.sciencebase.gov/auth"
        if env == "beta":
            self._graphql_url = "https://api-beta.staging.sciencebase.gov/graphql"
            self._realm = "ScienceBase-B"
        elif env == "dev":
            self._graphql_url = "http://localhost:4000/graphql"
            self._realm = "ScienceBase-B"
        else:
            self._graphql_url = "https://api.sciencebase.gov/graphql"
            self._realm = "ScienceBase"

    def get_graphql_url(self):
        '''get_graphql_url
        '''
        return self._graphql_url

    def login(self, username, password):
        """Log into ScienceBase using Keycloak

        :param username: The ScienceBase user to log in as
        :param password: The ScienceBase password for the given user
        :return: The SbSessionEx object with the user logged in
        """
        self._username = username

        try:
            authenticator = _keycloak_login(username, password, self._realm, self._auth_server_url)
            self._token = authenticator.get_access_token()
            self._refresh_token = authenticator.get_refresh_token()
            self._client_id = authenticator.keycloak_client_config.client_id
            self._token_server_uri = (
                authenticator.keycloak_client_config.get_token_server_uri()
            )
            self._token_expire = authenticator.get_token_expire()
            self._token_expire_refresh = authenticator.get_refresh_token_expire()
            self._token_expire_time = (
                self._token_expire + (datetime.today()).timestamp()
            )
            self._is_logged_in = True
        except Exception:
            self._logging.error(f"Keycloak login failed for {username} -- cloud services not available")
            self._is_logged_in = False

        return self

    def is_logged_in(self):
        '''is_logged_in
        '''
        return self._is_logged_in

    def get_current_user(self):
        '''get_current_user
        '''
        return self._username

    def get_logger(self):
        '''get_logger
        '''
        return self._logging

    def refresh_token(self):
        """Refresh tokens in ScienceBaseEx"""
        data = {
            "client_id": self._client_id,
            "grant_type": "refresh_token",
            "refresh_token": self._refresh_token,
        }
        token_resp = requests.post(self._token_server_uri, data=data)

        self._logging.info(token_resp.headers)
        self._logging.info(json.loads(token_resp.content.decode("utf-8")))

        if token_resp.status_code == 200:
            auth_token_json = token_resp.json()

            self._logging.info(auth_token_json)

            self._token = auth_token_json["access_token"]
            self._refresh_token = auth_token_json["refresh_token"]

            self._token_expire = auth_token_json["expires_in"]
            self._token_expire_refresh = auth_token_json["refresh_expires_in"]
            self._token_expire_time = (
                self._token_expire + (datetime.today()).timestamp()
            )
            self._logging.info("Token Refreshed.")
        else:
            raise Exception("Token Refreshed Failed.")

    def refresh_token_before_expire(self, refresh_amount):
        """Refresh token if token has not expired, but will expire with in some time,
        if token will expire with in that time then refresh will be triggered

        :refresh_amount: Amount subtracted (is seconds) from expired token value, that will trigger token refresh
        :return: True, if refresh is done, False, refresh is not triggered
        """
        current_time = (datetime.today()).timestamp() + refresh_amount

        if self._token_expire_time - current_time < 0:
            self.refresh_token()
            return True
        return False

    def refresh_token_time_remaining(self, refresh_amount):
        """Use for printing remaining time
        useful for debugging session timeout
        """
        current_time = (datetime.today()).timestamp() + refresh_amount
        return self._token_expire_time - current_time

    def upload_cloud_file_upload_session(self, item_id, filename, mimetype):
        '''upload_large_file_upload_session
        '''
        return client.upload_cloud_file_upload_session(item_id, filename, mimetype, self)

    def bulk_cloud_download(self, selected_rows):
        '''generate bulk cloud download tokenized links
        '''
        return client.bulk_cloud_download(selected_rows, self)

    def upload_s3_files(self, input):
        '''upload external S3 bucket files to ScienceBase Item
        '''
        return client.upload_s3_files(input, self)

    def unpublish_from_public_bucket(self, input):
        '''unpublish file from public S3 bucket
        '''
        return client.unpublish_from_public_bucket(input, self)

    def get_header(self):
        '''get_header
        '''
        return {
            "content-type": "application/json",
            "accept": "application/json",
            "authorization": "Bearer " + self._token,
        }

def _keycloak_login(username, password, realm, server_url):
    '''keycloak_login
    '''
    # Developer note: For some reason this method will not work inside the SbSessionEx class
    authenticator = auth.DirectAccessAuthenticator({
        "realm": realm,
        "auth-server-url": server_url,
        "ssl-required": "external",
        "resource": "sciencebasepy",
        "public-client": True,
        "confidential-port": 0,
    })
    authenticator.authenticate(username, password)

    return authenticator



import requests
import requests_mock
from sb3 import auth, client
import json
import pytest
import logging

adapter = requests_mock.Adapter()


@pytest.mark.usefixtures('unstub')
class SbSessionEx:
    def getGraphQLURL(self):
        return "https://api-beta.staging.sciencebase.gov/graphql"

    def get_header(self):
        return {
            "content-type": "application/json",
            "accept": "application/json",
            "authorization": f"Bearer AUTHORIZATION_TOKEN",
        }

    def getLogger(self):
        return logging


class TestAuthenticator:
    def test_upload(self, requests_mock, monkeypatch):

        session = SbSessionEx()

        itemId = "5f9982f1d34e9cff790bd2fe"
        filename = "sample_error.png"
        file_path = "tests/resources/sample_error.png"
        GRAPHQL_URL = "https://api-beta.staging.sciencebase.gov/graphql"

        upload_url_repsonse = """
            {
                "data": {
                    "createUploadSession": { 
                        "uploads": [
                            {
                                "url": "https://LONG_URL_WITH_PRESIGNED_URL"
                            }
                        ]
                    }
                }
            }
        """

        upload_response = """
            {
                "data": {
                    "item": { "id": "5f9982f1d34e9cff790bd2fe", "title": "himal-test" },
                    "createUploadSession": {
                    "uploads": [
                        {
                        "name": "sample_error15.png",
                        "url": "https://dev-is-sb-beta-content.s3.us-west-2.amazonaws.com/5f9982f1d34e9cff790bd2fe/sample_error15.png?Content-Type=image%2Fjpeg&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIARGHURBUVAHB66GAH%2F20201104%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20201104T165505Z&X-Amz-Expires=900&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEID%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCXVzLXdlc3QtMiJGMEQCIAkbmAecgCiYj6SyH3mt4ELsf4BlTOXNaxIOk%2BjJtGmbAiAtXMRpH0iu3b%2BWZiCxDqkiAGV0Zw8alrFdIVOlb4eQRSq2AwjZ%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAIaDDA4MjExNzIwMTE5NCIMWH8ITDo7PNTAXJ2ZKooD8ae44Sse8ZXiska%2FQQadDUicK6y5JOFLT4f3xGhecvzPYSDcaH3cLz6vt7P6n49DyAzkkfbnw%2F2tFlvs5f8cOh1%2Bu1%2BBmDnHPFRxHhQufIykNQdDgKQzgTVKhuoQLYWXKG1Nha6PFP4hQRd88DZiFDMDP67qYByqdjvJOGHrtXO9ornSq5JjxxSjaJ0EUjCldmvxshnxCXFhaxu8c5Al5IoMspO2IligOeYl2tkzwDSXUW9oKHke0IgYdaN30i%2BbAiusMzn6dgiahIpozesptVsOmgdcRgOuFUR5lJCTuhv12uwbQfr9BmKR6pt4RQWHDQspgCiHzphS9olba4TkyYJ3B%2FuI%2BK6%2FLSWYBNyKvqJdcg61L8zh31dyBskhPgbBrbPMf9PlLEHkFBJesIMamf1%2B2IA1RYr3AUAt%2BcGk6hd%2BZ%2FX%2BO%2ByhlFExIrH9NtjGwqRlAD59Pooit82rYicXhd9rxKlqBYaCnl9Owj0bWNkRGycrYzDIlnP4taVY2mdDVCTp4%2BPoq4ryOTCJp4v9BTrsAYTK1WAqtLIP5Mf86SvlRqNb%2FLkJNI%2FNeZR7%2FHlVpb2sfnJMrcolWctU%2BKj906K4SNY6%2B752IDCiln%2BIkk77Z4PyFXibHx9%2Bdo9SV9fd6jDdMRy1OfZ98RVeP0dJ1q5gFGKpNMWBY8xR7lN%2BPBYStg5ffwy%2BiOH8I3yrMr1R9Nw66L5p0vfLGmQ%2FFz7AXmM7IKSQjAFnWy3OwEl36yc2QKVDsobTJLoFjSG%2Fgn925JSIYa4NjExERdLzx54jDpnvneQPBiip3ERvUmm4nk%2BaEQ7J0POwA7t6DR6L6hPIUscZcMOR42V7AEExI918&X-Amz-Signature=c1bf1c669cedbd5a4948b00dae65dd33a16db0be7cf6a67899976cb6a669f8b3&X-Amz-SignedHeaders=host%3Bx-amz-acl&x-amz-acl=private"
                        }
                    ]
                    }
                }
            }
        """

        requests_mock.post(
            GRAPHQL_URL, json=json.loads(upload_url_repsonse), status_code=200
        )
        requests_mock.put(
            "https://LONG_URL_WITH_PRESIGNED_URL",
            json=json.loads(upload_response),
            status_code=200,
        )

        client.upload_test_with_graphql_upload_session(
            itemId, filename, file_path, session=session
        )



'''ScienceBase sbgraphql Client
'''
from pathlib import Path
import os
import requests
from progress.bar import Bar

from sb3 import querys

_CHUNK_SIZE = 104857600  # 104857600 == 100MB
_REFRESH_TOKEN_SUBTRACTED = 600  # 10 * 60

def upload_cloud_file_upload_session(itemid, file_path, mimetype, sb_session_ex):
    '''upload_cloud_file_upload_session
    :param itemid ID of the ScienceBase Item to which to upload the file
    :param filename File name
    :param filepath Full path to the file
    :param sbsession_ex SbSessionEx which has been logged in via Keycloak
    '''
    sb_session_ex.get_logger().info("upload_large_file_upload_session....")
    total_size = Path(file_path).stat().st_size
    total_chunks = int(total_size / _CHUNK_SIZE) + 1
    fpath = f'{itemid}/{os.path.basename(file_path)}'

    query_create_multi_part = querys.create_multipart_upload_session(fpath, mimetype, sb_session_ex.get_current_user())

    sb_session_ex.get_logger().info(query_create_multi_part)

    # Refresh token add amount to expire
    sb_session_ex.refresh_token_before_expire(_REFRESH_TOKEN_SUBTRACTED)
    requests_session = requests.session()

    sb_resp = requests_session.post(
        sb_session_ex.get_graphql_url(),
        headers=sb_session_ex.get_header(),
        json={"query": query_create_multi_part},
    )

    sb_session_ex.get_logger().info(
        f"get_item query response, status code: {sb_resp.status_code}"
    )

    if sb_resp.status_code == 200:
        sb_resp_json = sb_resp.json()
        sb_session_ex.get_logger().info(sb_resp_json)
    else:
        sb_resp_json = sb_resp.json()
        sb_session_ex.get_logger().error(sb_resp_json)
        raise Exception("Not status 200")

    unique_id = sb_resp_json["data"]["createMultipartUploadSession"]

    sb_session_ex.get_logger().info("unique_id : " + unique_id)

    part_number = 0
    parts_header = []

    sb_session_ex.get_logger().info("session: " + str(sb_session_ex))
    sb_session_ex.get_logger().info("totalChunks: " + str(total_chunks))

    prog_bar = Bar("Uploading", max=total_chunks)
    with open(file_path, "rb") as f:
        for piece in _read_in_chunks(f):
            part_number = part_number + 1

            # Refresh token add amount to expire
            sb_session_ex.refresh_token_before_expire(_REFRESH_TOKEN_SUBTRACTED)

            sb_session_ex.get_logger().info(
                "time remaining : "
                + str(sb_session_ex.refresh_token_time_remaining(_REFRESH_TOKEN_SUBTRACTED))
            )

            queryCreatePresignedUrlPart = querys.get_presigned_url_for_chunk(
                fpath, unique_id, part_number
            )
            sb_session_ex.get_logger().info(queryCreatePresignedUrlPart)

            sb_resp = requests_session.post(
                sb_session_ex.get_graphql_url(),
                headers=sb_session_ex.get_header(),
                json={"query": queryCreatePresignedUrlPart},
            )

            if sb_resp.status_code == 200:
                sb_resp_json = sb_resp.json()
                sb_session_ex.get_logger().info(sb_resp_json)
            else:
                sb_resp_json = sb_resp.json()
                sb_session_ex.get_logger().error(sb_resp_json)
                raise Exception("Not status 200")

            presignedUrl = sb_resp_json["data"]["getPreSignedUrlForChunk"]

            sb_session_ex.get_logger().info(presignedUrl)

            res = requests_session.put(presignedUrl, data=piece)

            if sb_resp.status_code != 200:
                raise Exception("Not status 200")

            eTag = res.headers["ETag"]
            parts_header.append({"ETag": eTag, "PartNumber": part_number})
            prog_bar.next()

            ##############################################################
            sb_session_ex.get_logger().info("+++++++++++++++++++++++++++++++++")
            sb_session_ex.get_logger().info(eTag)
            sb_session_ex.get_logger().info("+++++++++++++++++++++++++++++++++")
            ##############################################################
    prog_bar.finish()

    query_create_multi_part = querys.complete_multipart_upload(
        fpath, unique_id, parts_header
    )

    sb_session_ex.get_logger().info(query_create_multi_part)

    sb_session_ex.refresh_token_before_expire(_REFRESH_TOKEN_SUBTRACTED)

    sb_resp = requests_session.post(
        sb_session_ex.get_graphql_url(),
        headers=sb_session_ex.get_header(),
        json={"query": query_create_multi_part},
    )

    if sb_resp.status_code == 200:
        sb_resp_json = sb_resp.json()
        sb_session_ex.get_logger().info(sb_resp_json)
    else:
        sb_resp_json = sb_resp.json()
        sb_session_ex.get_logger().error(sb_resp_json)
        raise Exception("Not status 200")

    return sb_resp.json()


def _read_in_chunks(file_object, chunk_size=_CHUNK_SIZE):
    """Lazy function (generator) to read a file piece by piece.
    Default chunk size: 1000k."""
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


def bulk_cloud_download(selected_rows, sb_session_ex):
    query = """
          query getS3DownloadUrl($input: SaveFileInputs){
              getS3DownloadUrl(input: $input){
                downloadUri
              }
          }
        """

    variables = {"input": {"selectedRows": selected_rows}}

    requests_session = requests.session()

    sb_resp = requests_session.post(
                sb_session_ex.get_graphql_url(),
                headers=sb_session_ex.get_header(),
                json={'query': query, 'variables': variables}
            )

    if sb_resp.status_code != 200:
        sb_resp_json = sb_resp.json()
        sb_session_ex.get_logger().error(sb_resp_json)
        raise Exception("Not status 200")

    return sb_resp.json()


def unpublish_from_public_bucket(input, sb_session_ex):
    query = """
                mutation unpublishFile($input: UnpublishFileInput!){
                    unpublishFile(input: $input){
                      itemFile {
                        key
                      }
                    }
                }
            """

    variables = {"input": input}

    requests_session = requests.session()

    sb_resp = requests_session.post(
        sb_session_ex.get_graphql_url(),
        headers=sb_session_ex.get_header(),
        json={'query': query, 'variables': variables}
    )

    if sb_resp.status_code != 200:
        sb_resp_json = sb_resp.json()
        sb_session_ex.get_logger().error(sb_resp_json)
        raise Exception("Not status 200")

    return sb_resp.json()


def upload_s3_files(input, sb_session_ex):
    query = """
                mutation uploadS3Files($input: UploadS3FilesInput!){
                    uploadS3Files(input: $input){
                      id 
                    }
                }
            """

    variables = {"input": input}

    requests_session = requests.session()

    sb_resp = requests_session.post(
        sb_session_ex.get_graphql_url(),
        headers=sb_session_ex.get_header(),
        json={'query': query, 'variables': variables}
    )

    if sb_resp.status_code != 200 or 'errors' in sb_resp.json():
        sb_session_ex.get_logger().error(sb_resp.json())
        raise Exception("Not status 200")

    return sb_resp.json()



import requests
import requests_mock
from sb3 import auth
import json

adapter = requests_mock.Adapter()


class TestAuthenticator:
    def test_login(self, requests_mock):
        keycloak_client_config = {
            "realm": "ScienceBase-B",
            "auth-server-url": "https://www.sciencebase.gov/auth",
            "ssl-required": "external",
            "resource": "sb-steve-test-2",
            "public-client": True,
            "confidential-port": 0,
        }

        token_server_uri = (
            f"{keycloak_client_config['auth-server-url']}/realms/"
            f"{keycloak_client_config['realm']}/protocol/openid-connect/token"
        )

        return_json = """
        {
            "access_token": "REAL_LONG_ACCESS_TOKEN_STRING",
            "expires_in": 1800,
            "refresh_expires_in": 1800,
            "refresh_token": "REAL_LONG_REFRESH_TOKEN_STRING",
            "token_type": "bearer",
            "not-before-policy": 0,
            "session_state": "6f470ef1-d087-4fc1-9fe4-ab135dba02b4",
            "scope": "email profile"
        }
        """
        requests_mock.post(
            token_server_uri, json=json.loads(return_json), status_code=200
        )

        authenticator = auth.DirectAccessAuthenticator(keycloak_client_config)
        authenticator.authenticate("testuser", "testpassword")
        assert authenticator.get_access_token() == "REAL_LONG_ACCESS_TOKEN_STRING"


'''ScienceBase Keycloak Authentication Library
'''
import json
import requests

class KeycloakClientConfig:
    """Stores the Keycloak OIDC JSON with some properties pull out for convenience
    """
    _token_server_uri = None
    _auth_server_uri = None

    def __init__(self, oidc_config):
        """
        pass in either a dict from the JSON config or a a path to the JSON config
        :param oidc_config: dict of the loaded JSON config
        """
        self.oidc_config = oidc_config

        self.client_id = oidc_config["resource"]
        self._auth_server_uri = (
            f"{oidc_config['auth-server-url']}/realms/"
            f"{oidc_config['realm']}/protocol/openid-connect/auth"
        )
        self._token_server_uri = (
            f"{oidc_config['auth-server-url']}/realms/"
            f"{oidc_config['realm']}/protocol/openid-connect/token"
        )

    def get_token_server_uri(self):
        '''get_token_server_uri
        '''
        return self._token_server_uri

    def get_auth_server_uri(self):
        '''get_auth_server_uri
        '''
        return self._auth_server_uri

class DirectAccessAuthenticator:
    '''DirectAccessAuthenticator
    Authenticates with the Keycloak server using Direct Access
    '''
    def __init__(self, keycloak_config):
        if isinstance(keycloak_config, KeycloakClientConfig):
            self.keycloak_client_config = keycloak_config
        elif isinstance(keycloak_config, dict):
            self.keycloak_client_config = KeycloakClientConfig(keycloak_config)
        else:
            raise ValueError(
                "keycloak_config must be an instance of KeycloakClientConfig or dict of the config data"
            )

        self.auth_token = None

    def get_access_token(self):
        '''get_access_token
        '''
        if self.auth_token is None:
            raise ValueError(
                "auth_token is missing, authenticate() must run successfully "
                "before calling get_access_token"
            )
        return self.auth_token["access_token"]

    def get_refresh_token(self):
        '''get_refresh_token
        '''
        if self.auth_token is None:
            raise ValueError(
                "auth_token is missing, authenticate() must run successfully "
                "before calling get_refresh_token"
            )
        return self.auth_token["refresh_token"]

    def get_token_expire(self):
        '''get_token_expire
        '''
        if self.auth_token is None:
            raise ValueError(
                "auth_token is missing, authenticate() must run successfully "
                "before calling get_refresh_token"
            )
        return self.auth_token["expires_in"]

    def get_refresh_token_expire(self):
        '''get_refresh_token_expire
        '''
        if self.auth_token is None:
            raise ValueError(
                "auth_token is missing, authenticate() must run successfully "
                "before calling get_refresh_token"
            )
        return self.auth_token["refresh_expires_in"]

    def authenticate(self, username=None, password=None):
        '''authenticate
        '''

        payload = {
            "client_id": self.keycloak_client_config.client_id,
            "grant_type": "password",
            "username": username,
            "password": password,
        }
        token_resp = requests.post(
            self.keycloak_client_config.get_token_server_uri(), data=payload
        )
        token_resp_json = token_resp.json()

        if token_resp.status_code != 200:
            raise Exception("Authentication Failed")

        self.auth_token = token_resp_json

    def __str__(self):
        return json.dumps(self.auth_token)



'''GraphQL Queries for interaction with ScienceBase Manager
'''
def create_multipart_upload_session(s3_filepath, content_type, username):
    '''create_multipart_upload_session
    '''
    return f'''
        query {{
            createMultipartUploadSession(
                object: "{s3_filepath}"
                contentType: "{content_type}"
                username: "{username}"               
            )
        }}
    '''

def get_presigned_url_for_chunk(s3_filepath, upload_id, part_number):
    '''get_presigned_url_for_chunk
    '''
    return f'''
        query {{
            getPreSignedUrlForChunk(object: "{s3_filepath}", upload_id: "{upload_id}", part_number: "{part_number}")
        }}
    '''

def complete_multipart_upload(item_str, upload_id, etag_payload):
    '''complete_multipart_upload
    '''
    etag_payload_array = str(etag_payload).replace("'","")

    return f'''
        query {{
            completeMultiPartUpload(
                    object: "{item_str}"
                    upload_id: "{upload_id}"
                    parts_eTags: {etag_payload_array}
            )
        }}
        
    '''
